import os
import MySQLdb
import ConfigParser
from jtalks.util.Logger import Logger


class DB:
  """
    A class to work with DB (upload, download, search). Note, that for all the operations we need a connection to
   database which created by connect_to_database method. To connection need name of environments, which contained config file with properties to database.
  """
  env = None
  config = None
  dbHost = None
  dbUser = None
  dbPass = None
  dbName = None
  dbDefiner = None
  cursor = None
  connection = None
  jcName = None
  jcDescription = None
  jcUrl = None
  jcNotify = None
  poulpeAdminPass = None
  logger = Logger("DB")

  def __init__(self, env):
    self.env = env
    self.config = ConfigParser.ConfigParser()
    self.config.read("configs/" + env + "/db.cfg")
    self.dbHost = self.config.get('db', 'host')
    self.dbUser = self.config.get('db', 'user')
    self.dbPass = self.config.get('db', 'pass')
    self.dbName = self.config.get('db', 'name')
    self.dbDefiner = self.config.get('db', 'definer')
    self.jcName = self.config.get('properties', 'jc_name')
    self.jcDescription = self.config.get('properties', 'jc_description')
    self.jcUrl = self.config.get('properties', 'jc_url')
    self.jcNotify = self.config.get('properties', 'jc_notify')
    self.poulpeAdminPass = self.config.get('properties', 'poulpe_admin_pass')

  def connect_to_database(self):
    """
      Create connection to database.
    """
    self.connection = MySQLdb.connect(host=self.dbHost, user=self.dbUser, passwd=self.dbPass)
    self.cursor = self.connection.cursor()

  def close_connection(self):
    """
      Close connection to database.
    """
    self.connection.close()

  def recreate_database(self):
    """
      Method removed database (whith name, which contains in dbName), and create new(empty)
      database whita same name.
    """
    self.logger.info("Recreating database {0}", self.dbName)
    self.cursor.execute('DROP DATABASE IF EXISTS ' + self.dbName)
    self.cursor.execute('CREATE DATABASE ' + self.dbName)


  def restore_database_from_file(self, backupPath):
    """
      This method get path to filename (.sql) with backup of database. And restoring backup to database whith name,
      which contains in dbName.
    """
    self.logger.info("Loading a db dump from [{0}]", backupPath)
    self.recreate_database()
    self.fix_definer_in_backup(backupPath)
    os.popen('mysql -u ' + self.dbUser + ' --password=' + self.dbPass + ' ' + self.dbName + ' < ' + backupPath).read()
    self.logger.info("Database restored: environment=[{0}]".format(self.env))


  def fix_definer_in_backup(self, backupPath):
    """
     For MySQL.
     If database contains VIEWS then created backup of database (by mysqldump) file contains lines with permissions (to
    user from origin database). When we restoring database to another server(it server don't have it user) we get a ERROR. To fix this problem, before restore we need replace all entries  (with DEFINER='{username}') to MySQL constant CURRENT_USER. This constant indicates that the law should give to the user on whose behalf is restored.
    """
    os.popen('sed -i \'s/DEFINER=' + self.dbDefiner + '/DEFINER=CURRENT_USER/g\' ' + backupPath).read()
    self.logger.info("Database definer fixed: environment=[{0}]".format(self.env))

  def update_properties_to_preprod(self):
    """
     This method update application properties in database. Need to call after restore.
    """
    self.cursor.execute('USE ' + self.dbName)
    self.logger.info("Changing forum name to [{0}]", self.jcName)
    self.cursor.execute(
      'UPDATE COMPONENTS SET NAME="' + self.jcName + '", DESCRIPTION="' + self.jcDescription + '" where COMPONENT_TYPE="FORUM"')
    self.logger.info("Switching off mail notifications")
    self.cursor.execute(
      'UPDATE PROPERTIES SET VALUE="' + self.jcNotify + '" where NAME="jcommune.sending_notifications_enabled"')
    self.cursor.execute('UPDATE PROPERTIES SET VALUE="' + self.jcUrl + '" where NAME="jcommune.url_address"')
    self.cursor.execute('UPDATE USERS SET PASSWORD=MD5(' + self.poulpeAdminPass + ') WHERE USERNAME="admin"')
    self.connection.commit()
    
   
  



import os
from jtalks.util.Logger import Logger

__author__ = 'ctapobep'


class EnvList:
  """
    Can list all the envs and apps.
  """
  logger = Logger("EnvList")

  def __init__(self, script_settings):
    self.script_settings = script_settings

  def list_envs(self):
    envs = os.listdir(self.script_settings.get_env_configs_dir())
    envs.remove("global-configuration.cfg")
    return self.logger.info('[%s]' % ', '.join(map(str, envs)))

  def list_projects(self, env):
    projects = os.listdir(os.path.join(self.script_settings.get_env_configs_dir(), env))
    return [project.replace(".cfg", "") for project in projects if project.endswith(".cfg")]

  def print_envs_with_projects(self):
    for env in self.list_envs():
      print env
      for project in self.list_projects(env):
        print "  |_" + project


import pkg_resources
import sys
from jtalks.util.Logger import Logger

__author__ = 'stanislav bashkirtsev'


class LibVersion:
  logger = Logger("LibVersions")

  def log_lib_versions(self):
    self.logger.info("python={0}", sys.version_info)
    self.__log_lib_version("requests")
    self.__log_lib_version("GitPython")
    self.__log_lib_version("mock")


  def __log_lib_version(self, libname):
    self.logger.info("{0}={1}", libname, pkg_resources.get_distribution(libname).version)


import os
import shutil
from git import GitCommandError
import git.repo as repo
from jtalks.util.Logger import Logger

__author__ = 'stanislav bashkirtsev'


class EnvironmentConfigGrabber:
  CLONE_REPO_TO = '/tmp/jtalks-cicd/environments'
  GRABBED_CONFIGS_LOCATION = CLONE_REPO_TO + "/configs"
  logger = Logger("EnvironmentConfigGrabber")

  def __init__(self, env_configs_root):
    self.env_configs_root = env_configs_root

  def grab_jtalks_configs(self):
    try:
      self.__remove_previous_git_folder__()
      repo.Repo.clone_from('git@jtalks.org:environments', self.CLONE_REPO_TO)
      self.__copy_grabbed_configs_into_work_dir__()
    except GitCommandError:
      self.logger.warn("You don't have access to JTalks repo with environment configs. You may want to use your own "
                       + "local environment. Create a folder with env name in configs directory. If you think you "
                       + "need access to JTalks internal envs (including UAT, DEV, PREPROD, PROD envs), "
                       + "you should write a request to project@jtalks.org.")

  def __remove_previous_git_folder__(self):
    try:
      self.logger.info("Removing {0} directory if it was there", self.CLONE_REPO_TO)
      shutil.rmtree(self.CLONE_REPO_TO)
    except OSError as e:
      if e.errno is not 2:#No such file or directory
        self.logger.warn(e.message)

  def __copy_grabbed_configs_into_work_dir__(self):
    grabbed_dirs_and_files = os.listdir(self.GRABBED_CONFIGS_LOCATION)
    for next_grabbed_file_or_dir in grabbed_dirs_and_files:
      destination_file = self.env_configs_root + next_grabbed_file_or_dir
      if os.path.exists(destination_file):
        self.logger.info("{0} will be overwritten by newer version from git repo", destination_file)
        self.__delete_file_or_dir__(destination_file)
      shutil.move(self.GRABBED_CONFIGS_LOCATION + "/" + next_grabbed_file_or_dir, destination_file)

  def __delete_file_or_dir__(self, destination_file):
    if os.path.isfile(destination_file):
      os.remove(destination_file)
    else:
      shutil.rmtree(destination_file)



from jtalks.ApplicationContext import ApplicationContext
from jtalks.util.Logger import Logger

__author__ = 'stanislav bashkirtsev'


class DeployToTomcatFacade:
  NEXUS_URL = "http://repo.jtalks.org/content/repositories/deployment-pipeline/deployment-pipeline/"
  logger = Logger("DeployToTomcatFacade")


  def deploy(self, build_number, project, env):
    if build_number is None:
      self.logger.error("Build number was not specified, see [{0}] to get list of builds", self.NEXUS_URL)
      raise RuntimeError
    if project not in ["jcommune", "poulpe"]:
      self.logger.error("A correct project should be specified: [poulpe, jcommune]")
      raise RuntimeError

    app_context = ApplicationContext(environment=env, project=project, build=build_number)
    app_context.script_settings.log_settings()
    app_context.environment_config_grabber().grab_jtalks_configs()
    app_context.nexus().download_war(project=app_context.script_settings.project)
    app_context.tomcat().deploy_war()
    app_context.sanity_test().check_app_started_correctly()



import os
from jtalks.Nexus import Nexus
from jtalks.Tomcat import Tomcat
from jtalks.backup.Backuper import Backuper
from jtalks.db.DbOperations import DbOperations
from jtalks.db.DbSettings import DbSettings
from jtalks.db.LoadDbFromBackup import LoadDbFromBackup
from jtalks.parser.TomcatServerXml import TomcatServerXml
from jtalks.sanity.SanityTest import SanityTest
from jtalks.settings.ScriptSettings import ScriptSettings
from jtalks.util.EnvList import EnvList
from jtalks.util.EnvironmentConfigGrabber import EnvironmentConfigGrabber

__author__ = 'stanislav bashkirtsev'


class ApplicationContext:
  """
    Carries similar ideas to Spring's AppContext - it builds all the objects. This IoC is manual though, the objects
    are all constructed manually in Python code.
  """

  def __init__(self, environment, project, build):
    self.script_settings = ScriptSettings(build=build, project=project, env=environment)
    self.script_settings.create_work_dir_if_absent()
    self.script_settings.log_settings()

  def nexus(self):
    return Nexus(build_number=self.script_settings.build)

  def tomcat(self):
    return Tomcat(backuper=self.backuper(), script_settings=self.script_settings)

  def backuper(self):
    return Backuper(self.script_settings.get_backup_folder(), self.script_settings, self.db_operations())

  def db_operations(self):
    return DbOperations(self.script_settings.env, self.db_settings())

  def db_settings(self):
    config_file_location = self.__project_config_file_location__(self.script_settings.env,
                                                                 self.script_settings.project)
    db_settings = DbSettings(project=self.script_settings.project, config_file_location=config_file_location)
    return db_settings

  def environment_config_grabber(self):
    return EnvironmentConfigGrabber(self.script_settings.get_env_configs_dir())

  def sanity_test(self):
    http_port = self.tomcat_server_xml().http_port()
    return SanityTest(tomcat_port=http_port, app_name=self.script_settings.get_app_final_name())

  def tomcat_server_xml(self):
    return TomcatServerXml.fromfile(self.script_settings.get_tomcat_location() + "/conf/server.xml")

  def load_db_from_backup(self):
    return LoadDbFromBackup(self.script_settings.env)

  def script_settings(self):
    return self.script_settings

  def env_list(self):
    return EnvList(script_settings=self.script_settings)

  def __project_config_file_location__(self, env, project):
    return os.path.join(self.script_settings.get_env_configs_dir(), env, project + ".xml")


import shutil
import subprocess
from subprocess import PIPE
import os

from jtalks.util.Logger import Logger


class Tomcat:
  """
  Class for deploying and backing up Tomcat applications
  """

  logger = Logger("Tomcat")

  def __init__(self, backuper, script_settings):
    self.backuper = backuper
    self.script_settings = script_settings

  def deploy_war(self):
    """
    Stops the Tomcat server, backups all necessary application data, copies
    new application data to Tomcat directories
    """
    self.logger.info("Deploying {0} to {1}", self.script_settings.project, self.script_settings.get_tomcat_location())
    self.stop()
    self.backuper.backup()
    self.move_war_to_webapps()
    self.put_configs_to_conf()
    self.start()

  def stop(self):
    """
    Stops the Tomcat server if it is running
    """
    stop_command = "pkill -9 -f {0}".format(self.script_settings.get_tomcat_location())
    self.logger.info("Killing tomcat [{0}]", stop_command)
    #dunno why but retcode always equals to SIGNAL (-9 in this case), didn't figure out how to 
    #distinguish errors from this
    retcode = subprocess.call([stop_command], shell=True, stdout=PIPE, stderr=PIPE)

  def move_war_to_webapps(self):
    """
    Moves application war-file to 'webapps' Tomcat subfolder
    """
    final_app_location = self.get_web_apps_location() + "/" + self.script_settings.get_app_final_name()
    self.remove_previous_app(final_app_location)
    self.logger.info("Putting new war file to Tomcat: [{0}]", final_app_location)
    shutil.move(self.script_settings.project + ".war", final_app_location + ".war")

  def remove_previous_app(self, app_location):
    if os.path.exists(app_location):
      self.logger.info("Removing previous app: [{0}]", app_location)
      shutil.rmtree(app_location)

    war_location = app_location + ".war"
    if os.path.exists(war_location):
      self.logger.info("Removing previous war file: [{0}]", war_location)
      os.remove(war_location)

  def get_config_file_location(self):
    return os.path.join(self.script_settings.get_env_configs_dir(), self.script_settings.env, self.get_config_name())

  def put_configs_to_conf(self):
    """
    Copies configuration files for application and ehcache to Tomcat directories
    """
    final_conf_location = self.get_config_folder_location() + "/" + self.script_settings.get_app_final_name() + ".xml"
    conf_file_location = self.get_config_file_location()
    ehcache_config_file_location = os.path.join(self.script_settings.get_env_configs_dir(), self.script_settings.env,
                                                self.get_ehcache_config_name())

    self.logger.info("Putting [{0}] into [{1}]", conf_file_location, final_conf_location)
    shutil.copyfile(conf_file_location, final_conf_location)
    shutil.copy(ehcache_config_file_location, self.script_settings.get_tomcat_location() + "/conf")

  def start(self):
    """
    Starts the Tomcat server
    """
    startup_file = self.script_settings.get_tomcat_location() + "/bin/startup.sh"
    self.logger.info("Starting Tomcat [{0}]", startup_file)
    subprocess.call(startup_file, shell=True, stdout=PIPE, stderr=PIPE)

  def get_ehcache_config_name(self):
    """
    Returns name of the Ehcache configuration file
    """
    return self.script_settings.project + ".ehcache.xml"

  def get_config_name(self):
    """
    Returns name of the Tomcat configuration file
    """
    return self.script_settings.project + ".xml"

  def get_config_folder_location(self):
    """
    Returns configuration folder for Tomcat
    """
    return os.path.join(self.script_settings.get_tomcat_location(), "conf", "Catalina", "localhost")

  def get_web_apps_location(self):
    """
    Returns path to web applications directory of Tomcat
    """
    return self.script_settings.get_tomcat_location() + "/webapps"


from xml.dom.minidom import parse
import re


class DbSettings:
  """
  Class keeping connection settings to the database
  """

  def __init__(self, project, config_file_location):
    """
    Creates connection settings object with given project name
    """
    self.dbHost = None
    self.dbUser = None
    self.dbPass = None
    self.dbName = None
    self.dbPort = None
    self.project = project.upper()
    self.parse_config(config_file_location)

  def parse_config(self, config_file_path):
    """
    Parses given config file and gets information about connection settings.
    """
    configs_doc = parse(config_file_path)
    env_elements = configs_doc.getElementsByTagName('Environment')

    for element in env_elements:
      name = element.getAttribute("name")
      value = element.getAttribute("value")

      if name == (self.project + "_DB_USER"):
        self.dbUser = value
      elif name == (self.project + "_DB_PASSWORD"):
        self.dbPass = value
      elif name == (self.project + "_DB_URL"):
        jdbc_pattern = 'mysql://(.*?):?(\d*)/(.*?)\?'
        (host, port, database) = re.compile(jdbc_pattern).findall(value)[0]
        self.dbHost = host
        self.dbName = database
        self.dbPort = port



from jtalks.DB import DB
from jtalks.SSH import SSH

__author__ = 'stanislav bashkirtsev'

class LoadDbFromBackup:
  def __init__(self, env):
    self.env = env

  def load(self):
    ssh = SSH(self.env)
    db = DB(self.env)
    ssh.download_backup_prod_db()
    db.connect_to_database()
    db.restore_database_from_file(ssh.sftpBackupFileName)
    db.update_properties_to_preprod()
    db.connection.close()
    ssh.remove_backup_prod_db_file()
